import argparse
import ipaddress
import os
import threading
from datetime import datetime
from queue import Queue


import logger
from Marks import Marks
from analysis import MainAnalyzer
from dpkt_sniffer import DpktSniffer
from scapy_sniffer import ScapySniffer

MAX_QUEUE_SIZE = 1000

packet_queue = Queue(maxsize=MAX_QUEUE_SIZE)


def print_help():
    print("Pro analyzování již zachyceného provozu se program spouští:\n"
          "main.py --pcapFile nazevPcapSouboru.pcapng --outputFile nazevVystupnihoSouboru.txt --filterName "
          "nazevVPNfiltru\n"
          "Pro analýzu aktuálního provozu na síti je třeba specifikovat síťové rozhraní, například:\n"
          "main.py --interface eth0 --outputFile nazevVystupnihoSouboru.txt --filterName nazevVPNfiltru\n"
          "Pro načtení konfigurace z konfiguračního souboru je třeba zadat název tohoto souboru:\n"
          "main.py --configFile nazevKonfiguračníhoSouboru.txt\n")


def get_mark_from_argument(filter_names: []):
    mark_dict = dict()
    mark_dict["IPSEC"] = Marks.IPSEC
    mark_dict["WIREGUARD"] = Marks.WIREGUARD
    mark_dict["OPENVPN"] = Marks.OPENVPN
    mark_dict["OPENVPN_MIKROTIK"] = Marks.OPENVPN_MIKROTIK
    try:
        return set([mark_dict[filter_name] for filter_name in filter_names])
    except KeyError:
        return -1


def arguments_parsing(parser):
    if parser is None:
        return -2

    exclusiveGroup = parser.add_mutually_exclusive_group(required=True)
    exclusiveGroup.add_argument('--pcap', metavar='<název pcapng souboru>', help="pcapng soubor ke zpracování")
    # &   exclusiveGroup.add_argument('--interface', metavar='<název síťového rozhraní>',
    # help="síťové rozhraní k analyzování probíhající komunikace") # TODO future
    parser.add_argument("--outputFile", metavar='<jméno vystupního souboru>',
                        help="název výstupního souboru s výsledky analýzy")
    parser.add_argument("--network", metavar='<analyzovaná síť>',
                        help="IPv4 síť, v jejímž rozsahu se budou hledat adresy komunikující přes tunel", required=True)
    parser.add_argument("--filterNames", metavar='<název VPN filtru>',
                        help="Dostupné filtry: IPSEC, OPENVPN, OPENVPN_MIKROTIK, WIREGUARD",
                        required=True)
    parser.add_argument("--library", metavar='<Název použité knihovny>', help="Dostupné knihovny: dpkt, Scapy",
                        required=False, default="dpkt")
    return parser.parse_args()


def get_results_csv(ip_addresses: dict):
    output = ""
    header = ["VPN", "ADDRESS_A", "ADDRESS_B", "PACKETS_A", "PACKETS_B"]
    output += ",".join(header) + "\n"
    packet_types = ["PACKETS_A", "PACKETS_B"]
    for vpn in ip_addresses:
        for address in ip_addresses[vpn]:
            split = address.split(";")
            output_line = [vpn.name, split[0], split[1]]
            for packet_type in packet_types:
                if packet_type in ip_addresses[vpn][address]:
                    output_line.append(str(ip_addresses[vpn][address][packet_type]))
                else:
                    output_line.append("0")
            output += ",".join(output_line) + "\n"

    if len(output) == len(header):
        return 0
    return output


def write_results(outputFile, results):
    if os.path.isfile(outputFile):
        print("ERROR: Soubor " + outputFileName + " již existuje! Zadejte jiný název výstupního souboru!\n")
    logger.write_to_log_file(results, outputFile)


# Hlavní funkce programu
# Probíhá zde kontrola argumentů a volání funkce pro zpracování síťového provozu
if __name__ == '__main__':
    t1 = datetime.now()
    parser = argparse.ArgumentParser(description='VPN analyzer')
    arguments = arguments_parsing(parser)

    try:
        if arguments.network is not None:
            try:
                ipaddress.ip_network(arguments.network)
            except ValueError:
                print("ERROR: Nesprávně zadaná adresa sítě! Program bude ukončen!\n")
                exit(-1)
        if arguments.filterNames is not None:

            filters = arguments.filterNames.upper().split(",")
            possibleMarks = get_mark_from_argument(filters)

            if possibleMarks == -1:
                print("ERROR: Zadaný VPN filtr neexistuje!\n")
                exit(-4)
            if arguments.pcap is not None:
                input_file = arguments.pcap
                if arguments.library == "scapy":
                    sniffer = ScapySniffer(packet_queue)
                else:
                    sniffer = DpktSniffer()
                analyzer = MainAnalyzer(possibleMarks, sniffer)
                threading.Thread(target=lambda: analyzer.start_analysis(arguments.network, packet_queue),
                                 daemon=True).start()
                sniffer.get_ipadresses(input_file, packet_queue)
                packet_queue.join()
                analyzer.post_processing()

                if arguments.outputFile is None:
                    outputFileName = "Detection.csv"
                else:
                    outputFileName = arguments.outputFile

                outputFileName_csv = outputFileName[:outputFileName.index(".")] + ".csv"
                results_csv = get_results_csv(analyzer.ip_addresses)
                if len(results_csv) == 0:
                    os.remove(outputFileName_csv)
                else:
                    write_results(outputFileName_csv, results_csv)

                t2 = datetime.now()
                t = t2 - t1
                print("Čas běhu: ", t)
            else:
                input_interface = arguments.interface
                # TODO future
    except OSError as e:
        print("ERROR: Zpracování či otevření pcap souboru se nezdařilo!\n" + str(e))
        exit(-1)
